local cp = function(msg)
	lib.log_error(msg, 2, "helium", "0.4.2 -indev")
end

local he_ver = "0.4.2 -indev"
local he_path = lib.get_path("helium", he_ver)

cp("Helium " .. he_ver .. " is operating out of " .. he_path)

local config = {
	async_process_time = gkini.ReadInt("helium", "async_process_time", 10),
}

local he = {
	preset = {},
	buttons = {},
	frames = {},
	async = {},
	--public container class
}

local tryfile = function(file)
	local skinfile = (IMAGE_DIR or "skins/platinum/") .. file
	return gksys.IsExist(skinfile) and skinfile or (he_path .. "img/" .. file)
end



local container_types = {
	["hbox"] = true,
	["vbox"] = true,
	["frame"] = true,
	["zbox"] = true,
	["cbox"] = true,
	["dialog"] = true,
	["sbox"] = true,
}

--Maps a dialog, and handles map_cb() for ALL CHILDREN
he.map_dialog = function(ihandle)
	local cb_children
	cb_children = function(ihandle)
		local children = {}
		while true do
			local obj = iup.GetNextChild(ihandle, children[#children])
			if not obj then 
				break
			elseif obj.map_cb and type(obj.map_cb) == "function" then
				obj:map_cb()
			end
			table.insert(children, obj)
			if container_types[iup.GetType(obj)] then
				cb_children(obj)
			end
		end
		children = nil
	end
	
	if iup.IsValid(ihandle) then
		if iup.GetType(ihandle) == "dialog" then
			ihandle:map()
		end
		cb_children(ihandle)
		iup.Refresh(ihandle)
	end
end

--[[ old he.index_container
creates a simple table mapping an iup object.
he.index_container = function(ihandle)
	assert(iup.IsValid(ihandle) == true, type(ihandle) .. ":" .. tostring(ihandle) .. " was not a valid iup object for helium.index_container")
	
	local index_object
	index_object = function(ihandle)
		local children = {}
		while true do
			local obj = iup.GetNextChild(ihandle, children[#children])
			local obj_for_table
			if not obj then
				break
			elseif containter_types[iup.GetType(obj)] then
				obj_for_table = index_object(obj)
				obj_for_table.type = iup.GetType(obj)
			else
				obj_for_table = obj
			end
			table.insert(children, obj_for_table)
		end
		
		return children
	end
	
	return index_object(ihandle)
end
]]--

--creates a simple table mapping an iup object. Replace iup.GetType with iup.GetClassName if using iup >3.xx
he.index_container = function(roothandle)
	local index_object
    index_object = function(ihandle)
        local children = {}
        local last_child = nil
        while true do
            local obj = iup.GetNextChild(ihandle, last_child)
            if not obj then
                break
            end
            local obj_type = iup.GetType(obj)
            local obj_for_table
            if container_types[obj_type] then
                obj_for_table = index_object(obj)
                obj_for_table["_type"] = obj_type
            else
                obj_for_table = obj
            end
            table.insert(children, obj_for_table)
            last_child = obj
        end
        return children
    end

    return index_object(roothandle)
end

--like Append, but as a prefix. puts an object on the front of an iup stack
he.iup_prepend = function(root, obj)
	assert(iup.IsValid(obj) and iup.IsValid(root), "helium.iup_prepend expects both root and object to be valid iup containers!")
	
	local contents = {}
	local next = iup.GetNextChild(root)
	
	while next do
		table.insert(contents, next)
		next = iup.GetNextChild(root, next)
	end
	
	for k, v in ipairs(contents) do
		v:detach()
	end
	
	root:append(obj)
	
	for k, v in ipairs(contents) do
		root:append(v)
	end
	
	iup.Refresh(root)
end

he.async.iup_prepend = function(root, obj, on_complete)
	assert(iup.IsValid(obj) and iup.IsValid(root), "he.iup_prepend_async expects both root and object to be valid IUP containers!")

	local timer = Timer()
	local children = {}
	local current_step = 1
	local start_time = gkmisc.GetGameTime()

	-- Determine chunk size based on async_process_time
	local chunk_size = math.max(1, math.floor(config.async_process_time * 0.10)) -- Scale percentage to reasonable chunks

	-- Step 1: Detach the object to prepare for insertion
	obj:detach()

	-- Step 2: Get all current children of the root
	local next_child = iup.GetNextChild(root)
	while next_child do
		table.insert(children, next_child)
		next_child = iup.GetNextChild(root, next_child)
	end

	-- Timer handler function
	local function handle_prepend()
		if current_step == 1 then
			-- Detach current children in chunks
			for i = 1, chunk_size do
				local child = table.remove(children, 1)
				if child then
					child:detach()
				else
					current_step = 2
					break
				end
			end
		elseif current_step == 2 then
			-- Append the new object
			root:append(obj)
			current_step = 3
		elseif current_step == 3 then
			-- Reattach children in chunks
			for i = 1, chunk_size do
				local child = table.remove(children, 1)
				if child then
					root:append(child)
				else
					current_step = 4
					break
				end
			end
		elseif current_step == 4 then
			-- Final step: Refresh the root
			iup.Refresh(root)
			timer:Stop()

			-- Invoke the callback if provided
			if type(on_complete) == "function" then
				on_complete()
			end

			-- Output the elapsed time for debugging
			--cp(string.format("Async prepend completed in %.2f seconds", gkmisc.GetGameTime() - start_time))
		end
	end

	-- Start the timer to run the handler function every 1 ms
	timer:SetTimeout(1, handle_prepend)
end


--inserts the element in the target position
he.iup_insert = function(root, ihandle, position)
	assert(iup.IsValid(root) and iup.IsValid(ihandle), "helium.iup_insert expects both root and object to be valid iup containers!")
	assert(type(position) == "number", "helium.iup_insert expects the insert position to be a number; was type " .. type(position))
	
	local contents = {}
	local next = iup.GetNextChild(root)
	
	while next do
		table.insert(contents, next)
		next = iup.GetNextChild(root, next)
	end
	
	for i=1, #contents do
		contents[i]:detach()
	end
	
	if position < 0 then
		position = #contents + position
	end
	
	if position == 0 then
		position = 1
	end
	
	if position > #contents then
		position = #contents
	end
	
	table.insert(contents, position, ihandle)
	
	for k, v in ipairs(contents) do
		root:append(v)
	end
	
	iup.Refresh(root)
end

--draugath's old iup.IsValid, kept for posterity
he.IsIup = function(ihandle)
	return pcall(iup.GetType, ihandle)
end

--gets pixel position of the mouse
he.get_mouse_abs_pos = function(x_off, y_off)
	local perX, perY = gkinterface.GetMousePosition()
	local absX = (gkinterface.GetXResolution() * perX) + (tonumber(x_off) or 0)
	local absY = (gkinterface.GetYResolution() * perY) + (tonumber(y_off) or 0)
	
	return absX, absY
end

--scales a 1d value so that sizes are similar on all systems
he.scale_size = function(expected_size, expected_default)
	assert(type(expected_size) == "number", "helium.scale_size expects a number for the 'expected size' (arg 1), got " .. type(expected_size))
	if (type(expected_default) ~= "number") or (expected_default < 1) then
		expected_default = 24
	end
	
	return (Font.Default / expected_default) * expected_size
end

--scales 2d sizes based on an expected font size
he.scale_2x = function(exp_x, exp_y, exp_def)
	if (type(exp_def) ~= "number") or (exp_def < 1) then
		exp_def = 24
	end
	return tostring(he.scale_size(exp_x, exp_def)) .. "x" .. tostring(he.scale_size(exp_y, exp_def))
end

local is_mobile = gkinterface.IsTouchModeEnabled() and 2 or 1
local mobile_scalar = function()
	--Font.Default when PC, Font.Default * 2 when mobile
	return Font.Default * is_mobile
end








--transparent frame template, for frame shenanigans.
he.frames.clearframe = function(intable)
	assert(type(intable) == "table", "Helium.clearframe expects a table for its argument, got a " .. type(intable))
	assert((iup.IsValid(intable[1])) or (intable[1] == nil), "Helium.solidframe input table did not have a valid IUP element at [1]; got " .. type(intable[1]))
	
	local defaults = {
		image = tryfile("solidframe.png"),
		bgcolor = "0 0 0 0 *",
		segmented = "0 0 1 1",
		iup.vbox { },
	}
	
	for k, v in pairs(intable) do
		defaults[k] = v
	end
	
	return iup.frame(defaults)
end

--very simple opaque panel
he.frames.solidframe = function(intable)
	assert(type(intable) == "table", "Helium.solidframe expects a table for its argument, got a " .. type(intable))
	assert((iup.IsValid(intable[1])) or (intable[1] == nil), "Helium.solidframe input table did not have a valid IUP element at [1]; got " .. type(intable[1]))
	
	local defaults = {
		image = tryfile("solidframe.png"),
		bgcolor = "255 255 255 255 *",
		segmented = "0.1 0.1 0.9 0.9",
		iup.vbox { },
	}
	
	for k, v in pairs(intable) do
		defaults[k] = v
	end
	
	return iup.frame(defaults)
end

--very simple edge frame
he.frames.borderframe = function(intable)
	assert(type(intable) == "table", "Helium.borderframe expects a table for its argument, got a " .. type(intable))
	assert((iup.IsValid(intable[1])) or (intable[1] == nil), "Helium.solidframe input table did not have a valid IUP element at [1]; got " .. type(intable[1]))
	
	local defaults = {
		image = tryfile("borderframe.png"),
		bgcolor = "255 255 255 255 *",
		segmented = "0.1 0.1 0.9 0.9",
		iup.vbox { },
	}
	
	for k, v in pairs(intable) do
		defaults[k] = v
	end
	
	return iup.frame(defaults)
end

--primitive object used for highlite rules
he.highlite_panel = function()
	local highlite_obj = iup.label {
		title = "",
		image = tryfile("buttonRounded.png"),
		bgcolor = "255 255 255 0 *",
		expand = "YES",
		highlite = function(self)
			self.bgcolor = "255 255 255 255 *"
		end,
		clear_highlite = function(self)
			self.bgcolor = "255 255 255 0 *"
		end,
	}
	
	return highlite_obj
end

--primitive spacer using a frame object, mimics html <hr> kinda
he.page_rule = function(intable)
	local default = {
		orientation = "HORIZONTAL",
	}
	
	local object = he.control.solidframe {
		iup.vbox {
			((default.orientation == "VERTICAL") or (default.orientation == "ALL")) and iup.fill { } or nil,
			iup.hbox {
				((default.orientation == "HORIZONTAL") or (default.orientation == "ALL")) and iup.fill { } or nil,
			},
		},
	}
	
	return object
end

--invisible button to cover selected object
he.coverbutton = function(intable)
	local default = {
		action = function(self) end,
		expand = "NO",
		highlite = "NO",
		[1] = iup.vbox { },
		size = nil,
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local hlpane = he.highlite_panel()
	
	local cover = iup.button {
		title = "", 
		bgcolor = "0 0 0 0 *",
		expand = "YES",
		action = default.action,
		size = default.size,
		enterwindow_cb = function(self)
			hlpane:highlite()
		end,
		leavewindow_cb = function(self)
			hlpane:clear_highlite()
		end,
		destroy_cb = function()
			hlpane:destroy()
		end,
	}
	
	return iup.zbox {
		all = "YES",
		expand = default.expand,
		alignment = "ACENTER",
		default.highlite == "YES" and hlpane or nil,
		default[1],
		cover,
	}
end

--selectable_text uses a cover_button on top of a label. eventually will use frame/image properties to underline text maybe?
he.select_text = function(intable)
	local default = {
		title = "selectable text",
		action = function(self) end,
		font = Font.Default,
		fgcolor = nil,
		bgcolor = nil,
		highlite = "NO",
		size = nil,
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local obj = he.coverbutton {
		action = default.action,
		highlite = default.highlite,
		size = default.size,
		iup.label {
			title = default.title,
			font = default.font,
			fgcolor = default.fgcolor,
			bgcolor = default.bgcolor,
		},
	}
	
	return obj
end

--preset he.control.select_text used for URLs
he.link_text = function(intable)
	local default = {
		title = "selectable url",
		url = "http://vendetta-online.com/",
		font = Font.Default,
		fgcolor = "128 0 255",
		bgcolor = nil,
		highlite = "YES",
		size = nil,
		link_image = tryfile("weblink.png"),
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local linkimg = iup.label {
		title = "",
		image = default.link_image,
		size = tostring(default.font) .. "x" .. tostring(default.font),
	}
	
	local obj = he.coverbutton {
		action = function(self)
			Game.OpenWebBrowser(default.url or default.title)
		end,
		highlite = default.highlite,
		size = default.size,
		iup.hbox {
			iup.label {
				title = default.title,
				font = default.font,
				fgcolor = default.fgcolor,
				bgcolor = default.bgcolor,
			},
			default.link_image and linkimg or nil,
		},
	}
	
	return obj
end

--a simple frame that can be expanded or shrank vertically; contents are stored and append/detached
he.vexpandbox = function(intable)
	local default = {
		state = "CLOSED",
		map_cb = nil,
		drawer_cb = function(self) end,
		[1] = iup.vbox {},
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local vexpand
	local contents = iup.vbox {}
	local contents_table = {}
	
	for k, v in ipairs(default) do 
		table.insert(contents_table, v)
		if default.state == "OPEN" then
			contents:append(v)
		end
	end
	
	vexpand = he.control.clearframe {
		map_cb = default.map_cb,
		contents,
		state = default.state,
		get_obj_table = function(self)
			return contents_table
		end,
		drawer_toggle = function(self, new_state)
			if new_state == "OPEN" or new_state == "CLOSED" then
				self.state = new_state
			else
				self.state = self.state == "OPEN" and "CLOSED" or "OPEN"
			end
			
			if self.state == "CLOSED" then
				for k, v in ipairs(contents_table) do
					v:detach()
				end
			else
				for k, v in ipairs(contents_table) do
					contents:append(v)
				end
			end
			
			iup.Refresh(self)
			default.drawer_cb(self)
		end,
	}
	
	return vexpand
end

--a simple frame that can be expanded or shrank horizontally; contents are stored and append/detached
he.hexpandbox = function()
	local default = {
		state = "CLOSED",
		map_cb = nil,
		drawer_cb = function(self) end,
		[1] = iup.hbox {},
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local vexpand
	local contents = iup.hbox {}
	local contents_table = {}
	
	for k, v in ipairs(default) do 
		table.insert(contents_table, v)
		if default.state == "OPEN" then
			contents:append(v)
		end
	end
	
	hexpand = he.control.clearframe {
		map_cb = default.map_cb,
		contents,
		state = default.state,
		get_obj_table = function(self)
			return contents_table
		end,
		drawer_toggle = function(self, new_state)
			if new_state == "OPEN" or new_state == "CLOSED" then
				self.state = new_state
			else
				self.state = self.state == "OPEN" and "CLOSED" or "OPEN"
			end
			
			if self.state == "CLOSED" then
				for k, v in ipairs(contents_table) do
					v:detach()
				end
			else
				for k, v in ipairs(contents_table) do
					contents:append(v)
				end
			end
			
			iup.Refresh(self)
			default.drawer_cb(self)
		end,
	}
	
	return hexpand
end

he.ticker = function(intable)
	local default = {
		constrain = "YES", --limit to min/max if yes
		min = 0,
		max = 10,
		default = 0,
		readonly = "NO", --prevents text edits. buttons are active.
		active = "YES", --overrides readonly; prevents any changes
		action = function(self, value, caller) end,
		v_size = tostring(mobile_scalar()),
		h_size = "200", --h_size of text control
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local tickframe
	
	local update_val = function(input, mod, editor)
		mod = tonumber(mod) or 0
		input = (tonumber(input) or 0) + mod
		if default.constrain == "YES" then
			if input > default.max then
				input = default.max
			elseif input < default.min then
				input = default.min
			end
		end
		tickframe.value = tostring(input)
		default.action(tickframe, tickframe.value, editor)
	end
	
	local tickreader = iup.text {
		size = tostring(default.h_size) .. "x" .. tostring(default.v_size),
		value = default.default,
		readonly = (default.active == "NO" and "YES") or default.readonly,
		action = function(self)
			self.value = string.match(self.value, "%d+%.?%d*")
			update_val(self.value, 0, "edit")
			self.value = tickframe.value
		end,
	}
	
	local uptick = iup.stationbutton {
		title = "+",
		size = tostring(default.v_size) .. "x" .. tostring(default.v_size / 2),
		active = default.active,
		action = function()
			update_val(tickframe.value, 1, "up")
			tickreader.value = tickframe.value
		end,
	}
	
	local downtick = iup.stationbutton {
		title = "-",
		size = tostring(default.v_size) .. "x" .. tostring(default.v_size / 2),
		active = default.active,
		action = function()
			update_val(tickframe.value, -1, "down")
			tickreader.value = tickframe.value
		end,
	}
	
	tickframe = iup.stationsubframe_nomargin {
		iup.hbox {
			tickreader,
			iup.vbox {
				uptick,
				downtick,
			},
		},
		value = default.default,
	}
	
	return tickframe
end


he.hslider = function(intable)
	
	local default = {
		xmin = 0,
		xmax = 100,
		dx = 30,
		posx = 0,
		scrollbar = "HORIZONTAL",
		expand = "HORIZONTAL",
		scroll_cb = function(self) end,
		border = "NO",
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local scroller = iup.canvas {
		scrollbar = default.scrollbar,
		expand = default.expand,
		border = default.border,
		
		xmin = default.xmin,
		xmax = default.xmax,
		dx = default.dx,
		posx = default.posx,
		ppos = -1, --% position value, read only
		
		scroll_cb = function(self, ...)
			local curpos = self.posx
			local min = self.xmin
			local max = self.xmax
			self.ppos = (curpos / (max - min) * 100)
			default.scroll_cb(self, ...)
		end,
		set_range = function(self, min, max)
			default.min = min or default.xmin
			default.max = max or default.xmax
		end,
		get_range = function()
			return default.min, default.max
		end,
		get_pos = function(self)
			local range = default.xmax - default.xmin
			return (default.xpos / 100) * range
		end,
		get_current_pos = function(self)
			return default.xpos
		end,
		map_cb = function(self)
			local container_w = iup.GetParent(self).w
			container_w = container_w or 100
			default.xmax = container_w
			iup.Refresh(self)
		end,
	}
	
	return scroller
end


he.vslider = function(intable)
	
	local default = {
		ymin = 0,
		ymax = 100,
		dy = 30,
		posy = 0,
		scrollbar = "VERTICAL",
		expand = "VERTICAL",
		scroll_cb = function(self) end,
		border = "NO",
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local scroller = iup.canvas {
		scrollbar = default.scrollbar,
		expand = default.expand,
		border = default.border,
		
		ymin = default.ymin,
		ymax = default.ymax,
		dy = default.dy,
		posy = default.posy,
		ppos = -1, --% position value, read only
		
		scroll_cb = function(self)
			self.ppos = (self.posy / (self.ymax - self.ymin) * 100)
			default.scroll_cb(self)
		end,
		get_pos = function(self)
			self.ppos = (self.posy / (self.ymax - self.ymin) * 100)
			return self.ppos
		end,
		set_pos = function(self, new_ppos)
			local tposy = (new_ppos * (self.ymax - self.ymin)) / 100
			if tposy < tonumber(self.ymin) then
				tposy = tonumber(self.ymin)
			elseif tposy > tonumber(self.ymax) then
				tposy = tonumber(self.ymax)
			end
			self.posy = tposy
			iup.Refresh(self)
			self.scroll_cb(self)
		end,
		get_current_pos = function(self)
			return self.ypos
		end,
		set_current_pos = function(self, new_pos)
			self.ypos = new_pos
		end,
		map_cb = function(self)
			if self.expand == "VERTICAL" then
				local container_h = iup.GetParent(self).h
				container_h = container_h or 100
				default.ymax = container_h
				iup.Refresh(self)
			end
		end,
	}
	
	return scroller
end





he.vscroll = function(intable)
	--creates a vertically scrolling content list, without using iup.list as a control menu
	local default = {
		expand = "NO",
		fit_parent = "YES", --expand to parent frame's size (more versatile than expand, but requires parent to be a frame or other object with defined size (not a vbox/hbox/etc))
		passed_data = {}, --functions to pass into final iup object
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local content_list = {}
	local content_vbox = iup.vbox {}
	
	for k, v in ipairs(default) do
		if iup.IsValid(v) then
			table.insert(content_list, v)
			iup.Append(content_vbox, v)
		end
	end
	
	local content_frame = he.frames.clearframe {
		cx = 0,
		cy = 0,
		content_vbox,
	}
	
	local scroller
	local scroll_timer = Timer()
		--a timer is used because scroll sliders can lag the game; scroll_cb called EVERY FRAME while the user is dragging it
	local scroll_update_func = function()
		local self = scroller
		local cmax = content_frame.h
		content_frame.cy = ((self.posy / 100) * (cmax - self.h)) * -1
		iup.Refresh(content_frame)
	end
	
	scroller = he.vslider {
		scroll_cb = function(self)
			scroll_timer:SetTimeout(1, scroll_update_func)
		end,
	}
	
	local cbox_area = iup.cbox {
		expand = "YES",
		content_frame,
	}
	
	local root_frame = he.frames.clearframe {
		size = default.size,
		expand = default.expand,
		iup.hbox {
			cbox_area,
			scroller,
		},
		map_cb = function(self)
			if default.fit_parent == "YES" then
				local root = iup.GetParent(self)
				local w = root.w
				local h = root.h
				--self.size = tostring(w) .. "x" .. tostring(h)
				cbox_area.size = tostring(w - Font.Default) .. "x" .. tostring(h)
				scroller.size = tostring(Font.Default) .. "x" .. tostring(h)
				console_print("w: " .. tostring(w))
				console_print("h: " .. tostring(h))
				console_print("size: " .. tostring(self.size))
				iup.Refresh(self)
			end
		end,
		get_list = function()
			return content_list
		end,
		detach_entry = function(self, index)
			if true then
				return
			end
			if not index then
				index = #content_list
			end
			content_vbox:detach(content_list[index])
			iup.Refresh(self)
			return table.remove(content_list, index)
		end,
		clear_all = function(self)
			local counter = #content_list
			while counter > 0 do
				content_vbox:detach(content_list[counter])
				table.remove(content_list, index)
				counter = counter - 1
			end
			iup.Refresh(self)
		end,
		append_entry = function(self, new_content)
			local temproot = iup.GetParent(cbox_area)
			cbox_area:detach()
			table.insert(content_list, new_content)
			content_vbox:append(new_content)
			--he.iup_prepend(temproot, cbox_area)
			he.iup_prepend(temproot, cbox_area)
			--he.map_dialog(new_content)
			iup.Refresh(content_vbox)
		end,
		insert_entry = function(self, new_content, index)
			table.insert(content_list, index, new_content)
			he.insert(content_vbox, new_content, index)
			iup.Refresh(self)
		end,
		scroll_to = function(self, percent_scroll)
			scroller:set_pos(percent_scroll)
		end,
	}
	
	for k, v in pairs(default.passed_data) do
		root_frame[k] = v
	end
	
	return root_frame
end

he.hscroll = function(intable)
	--creates a horizontal scrolling content list, which iup.list is incapable of doing
	local default = {
		expand = "NO",
		fit_parent = "YES", --expand w to parent frame's size
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local content_list = {}
	local content_hbox = iup.hbox {}
	
	for k, v in ipairs(default) do
		if iup.IsValid(v) then
			table.insert(content_list, v)
			iup.Append(content_hbox, v)
		end
	end
	
	local content_frame = he.frames.clearframe {
		expand = "HORIZONTAL",
		cx = 0,
		cy = 0,
		content_hbox,
	}
	
	local scroller = he.hslider {
		scroll_cb = function(self)
			local cmax = content_frame.w
			content_frame.cx = ((self.posx / 100) * (cmax - self.w)) * -1
			iup.Refresh(content_frame)
		end,
	}
	
	local cbox_area = iup.cbox {
		expand = "HORIZONTAL",
		content_frame,
	}
	
	local root_frame = he.frames.clearframe {
		size = default.size,
		expand = default.expand,
		iup.vbox {
			cbox_area,
			scroller,
		},
		map_cb = function(self)
			if default.fit_parent == "YES" then
				cbox_area.size = tostring(self.w - Font.Default) .. "x"
				scroller.size = tostring(self.w) .. "x"
				iup.Refresh(self)
			end
		end,
		get_list = function()
			return content_list
		end,
		detach_entry = function(self, index)
			if not index then
				index = #content_list
			end
			content_hbox:detach(content_list[index])
			return table.remove(content_list, index)
		end,
		clear_all = function(self)
			local counter = #content_list
			while counter > 0 do
				content_hbox:detach(content_list[counter])
				table.remove(content_list, index)
				counter = counter - 1
			end
		end,
		append_entry = function(self, new_content)
			table.insert(content_list, new_content)
			content_hbox:append(new_content)
		end,
		insert_entry = function(self, new_content, index)
			table.insert(content_list, index, new_content)
			he.insert(content_hbox, new_content, index)
		end,
		
	}
	
	return root_frame
end

he.scrolling_tablist = function(intable)
	--modified hscroll specifically for creating a button row
	local default = {
		tab_select_cb = function(self_container, index)
			--called when a valid tab is selected.
		end,
		fit_parent = "YES",
		active_color = "255 255 255",
		inactive_color = "128 128 128",
		default_tab = 0,
		"Tab 1",
		
		--[[
			eventually, add keys like "button_provider" here that let the library generate tabs not based on stationbutton.
		]]--
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local tablist
	
	local tab_list = {}
	local tab_cont = iup.hbox {}
	
	for k, v in ipairs(default) do
		local tab = iup.stationbutton {
			title = tostring(v),
			_index = k,
			bgcolor = k == default.default_tab and default.active_color or default.inactive_color,
			action = function(self)
				for _, button in ipairs(tab_list) do
					button.bgcolor = default.inactive_color
				end
				self.bgcolor = default.active_color
				default.tab_select_cb(tablist, self._index)
			end,
		}
		
		table.insert(tab_list, tab)
		tab_cont:append(tab)
		
	end
	
	local tab_frame = he.frames.clearframe {
		--expand = "HORIZONTAL",
		cx = 0,
		cy = 0,
		tab_cont,
	}
	
	local held = false
	local scroller
	local scroll_timer = Timer()
	local scroll_update_func = function()
		if held then
			held = false
		end
		local self = scroller
		local cmax = tab_frame.w
		tab_frame.cx = ((self.posx / 100) * (cmax - self.w)) * -1
		iup.Refresh(tab_frame)
	end
	
	scroller = he.hslider {
		expand = "HORIZONTAL",
		scroll_cb = function(self)
			if held then
				
			else
				scroll_timer:SetTimeout(1, scroll_update_func)
				held = true
			end
		end,
	}
	
	local cbox_area = iup.cbox {
		expand = "HORIZONTAL",
		tab_frame,
	}
	
	tablist = he.frames.clearframe {
		expand = "HORIZONTAL",
		iup.vbox {
			iup.hbox {
				iup.fill { },
			},
			cbox_area,
			scroller,
		},
		get_tab_title = function(self, index)
			index = tonumber(index) or -1
			return (tab_list[index] and tab_list[index].title) or nil
		end,
		map_cb = function(self)
			if default.fit_parent == "YES" then
				local root = iup.GetParent(self)
				print("Parent: " .. tostring(root))
				print("	" .. tostring(iup.GetType(root)))
				print("	" .. tostring(root.size))
				print("	" .. tostring(iup.GetName(root)))
				local w = tonumber(root.w) or 0
				local h = tonumber(root.h) or 0
				--self.size = tostring(w) .. "x" .. tostring(h)
				--tab_frame.size = tostring(w) .. "x"
				if w < 100 then
					w = self.manual_wid or default.manual_wid or (gkinterface.GetXResolution() / 2)
				end
				--scroller.size = tostring(w) .. "x"
				iup.Refresh(self)
				iup.Refresh(tab_frame)
				cbox_area.size = scroller.size
				iup.Refresh(cbox_area)
				print("tab_frame size: " .. tostring(tab_frame.size))
				print("cbox_area size: " .. tostring(cbox_area.size))
				print("size: " .. tostring(self.size))
				print("w given was " .. tostring(w))
			elseif default.size then
				
			end
		end,
	}
	
	return tablist
end


















--create a dialog at the position or cursor; adjusted to fit in game window
he.context_container = function(intable)
	--[[
		clickthrough:
			not registered or close: close the dialog
			ignore: do nothing
		show:
			YES: Show dialog immediately
			NO: Dialog remains hidden until called
		alignment:
			defines the orientation of the dialog from the position provided
	]]--
	local mx, my = he.get_mouse_abs_pos()
	
	local default = {
		clickthrough = "CLOSE",
		action = function(self, action) end,
		show = "YES",
		alignment = "SE",
		pos_x = mx,
		pos_y = my,
		[1] = he.frames.solidframe {},
	}
	
	for i, v in pairs(intable) do
		default[i] = v
	end
	
	local obj_container = he.frames.clearframe {
		default[1],
		cx = default[1].cx,
		cy = default[1].cy,
	}
	
	local root_diag = iup.dialog {
		topmost = "YES",
		fullscreen = "YES",
		bgcolor = "0 0 0 0 *",
		map_cb = function(self)
			local offset_x = 0
			local offset_y = 0
			local res_x = gkinterface.GetXResolution()
			local res_y = gkinterface.GetYResolution()
			local diag_pos = {} --this gets modified
			local diag_size = {} --this doesn't get modified
			for value in string.gmatch(obj_container.size, "%d+") do
				table.insert(diag_pos, tonumber(value))
				table.insert(diag_size, tonumber(value))
			end
			
			--set the dialog's start position
			diag_pos[1] = diag_pos[1] + default.pos_x
			diag_pos[2] = diag_pos[2] + default.pos_y
			
			--set the dialog's orientation from its start position
			if default.alignment == "ACENTER" then
				diag_pos[1] = diag_pos[1] - (diag_size[1] / 2)
				diag_pos[2] = diag_pos[2] - (diag_size[2] / 2)
			elseif default.alignment == "SE" then
				--no change needed
			elseif default.alignment == "SW" then
				diag_pos[1] = diag_pos[1] - diag_size[1]
			elseif default.alignment == "NE" then
				diag_pos[2] = diag_pos[2] - diag_size[2]
			elseif default.alignment == "NW" then
				diag_pos[1] = diag_pos[1] - diag_size[1]
				diag_pos[2] = diag_pos[2] - diag_size[2]
			end
			
			--fix dialog positioning over interface borders
			if diag_pos[1] < 0 then
				--move to screen border
				offset_x = diag_pos[1] * -1
			elseif diag_pos[1] > res_x then
				--subtract by difference in position and screen edge
				offset_x = diag_pos[1] - res_x
			end
			
			if diag_pos[2] < 0 then
				offset_y = diag_pos[2] * -1
			elseif diag_pos[2] > res_y then
				offset_y = diag_pos[2] - res_y
			end
			
			
			obj_container.cx = default.pos_x - offset_x
			obj_container.cy = default.pos_y - offset_y
			
			iup.Refresh(obj_container)
		end,
		iup.cbox {
			iup.canvas {
				button_cb = function(self, ...)
					if default.clickthrough ~= "IGNORE" then
						local args = {...}
						if args[2] == 1 then
							HideDialog(iup.GetDialog(self))
							default:action("_cancel")
						else
							
						end
					end
				end,
				border = "NO",
				size = HUDSize(1, 1),
				cx = 0,
				cy = 0,
			},
			obj_container,
		},
	}
	he.map_dialog(root_diag)
	
	if default.show == "YES" then
		ShowDialog(root_diag)
	end
	
	return root_diag
end


--uses he.context_container to create a centered subdialog
he.subdialog = function(intable)
	local default = {
		[1] = iup.vbox {}, --content to replace
		show = "YES",
		show_close = "YES",
		show_title = "YES",
		title = "Helium subdialog",
		bgcolor = nil,
		bgimage = nil,
		font = mobile_scalar(),
		action = function(self, event) end,
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local subframe_contents = {
		bgcolor = default.bgcolor,
		iup.hbox {
			iup.fill {
				size = "%1",
			},
			default[1], --cannot do this! Append later, or breaks on using image
			iup.fill {
				size = "%1",
			},
		},
	}
	
	local subdiag = iup.vbox {
		iup.hbox {
			iup.label {
				title = default.title,
				font = default.font,
			},
			--iup.fill { }, --need to make this shrink-able
			iup.stationbutton {
				title = "Close",
				visible = default.show_close,
				active = default.show_close,
				action = function(self)
					HideDialog(iup.GetDialog(self))
					default.action(self, "_close")
				end,
			},
		},
		default.bgimage and he.bgframe(subframe_contents) or iup.stationdoubleframe(subframe_contents),
	}
	
	return he.context_container {
		bgcolor = default.bgcolor,
		bgimage = default.bgimage,
		subdiag,
		show = default.show,
		alignment = "ACENTER",
		clickthrough = default.clickthrough,
		pos_x = gkinterface.GetXResolution() / 2,
		pos_y = gkinterface.GetYResolution() / 2,
		action = default.action,
	}
end

--create a button menu at the position or cursor
he.context_menu = function(intable)
	local default = {
		pos_x = nil, --x position override
		pos_y = nil, --y position override
		select_cb = function(selected_index, selected_text)
			
		end,
		font = Font.H6,
		show = "YES",
		"Option 1",
	}
	
	for k, v in pairs(intable) do
		default[k] = v
	end
	
	local button_table = {}
	local button_container = iup.vbox {}
	for k, v in ipairs(default) do
		local btn_obj = he.select_text {
			title = tostring(v),
			action = function(self)
				HideDialog(iup.GetDialog(self))
				default.select_cb(k, v)
			end,
			highlite = "YES",
			font = default.font,
			size = "x" .. tostring(default.font),
		}
		table.insert(button_table, btn_obj)
		iup.Append(button_container, btn_obj)
	end
	
	local context = he.context_container {
		show = default.show,
		action = function(self, status)
			default.select_cb(0, status)
		end,
		he.frames.solidframe {
			button_container,
		},
	}
	
	return context
end





lib.set_class("helium", he_ver, he)