-- renoise
-- pack-based SFX replacer

local cp = function(print_string)
	lib.log_error("[Renoise] " .. print_string, 1, "renoise", "0")
end

local cerr = function(err_string)
	lib.log_error("[Renoise] " .. err_string, 3, "renoise", "0")
	return false, err_string
end

local gkrs = gkini.ReadString

local rini = function(header, item, ini_path)
	return gkini.ReadString2(header, item, "null", ini_path)
end










local public = {}
local config = {}

local pack_registry = {
	--[[
		[id] = {
			path = 'path/to/pack.ini',
			name = pack_name,
			load = YES/NO
				if yes, loads this soundpack on startup
		}
		[path/to/pack.ini] = id,
		...
		...
		...
	]]--
}


--babel handle

local bstr = function(id, val)
	return val
end

local update_class, babel, lang_key
local babel_func = function()
	babel = lib.get_class("babel", "0")
	lang_key = babel.register(lib.get_path("renoise", "0") .. "lang/", {'en', 'es', 'fr', 'pt'})
	
	bstr = function(id, val)
		return babel.fetch(lang_key, id, val)
	end
	
	public.add_translation = function(path, lang_code)
		babel.add_new_lang(ref_id, path, lang_code)
	end
	
	update_class()
end



--class

update_class = function()
	local class = {
		CCD1 = true,
		open = nil,
		config = nil,
		smart_config = {
			title = "Renoise core",
			action = function(id, val)
				if id == "update_check" then
					Game.OpenWebBrowser("https://www.nexusmods.com/vendettaonline/mods/24")
				end
			end,
			"no_values1",
			"no_values2",
			"no_values3",
			"spacer",
			"update_check",
			spacer = {
				type = "spacer",
			},
			no_values1 = {
				type = "text",
				display = "The core module of Renoise does not support configuration.",
				align = "center",
			},
			no_values2 = {
				type = "text",
				display = "Please use a front-end interface instead.",
				align = "center",
			},
			no_values3 = {
				type = "text",
				display = "rngui is provided with Renoise core for your convenience.",
				align = "center",
			},
			update_check = {
				type = "action",
				display = "",
				align = "right",
				[1] = bstr(3, "Check for updates") .. "...",
			},
		},
		description = bstr(1, "Renoise is a core module facilitating custom sound effect packs to be used in Vendetta Online"),
		commands = {},
		manifest = {},
	}
	
	for k, v in pairs(class) do
		public[k] = v
	end
	
	lib.set_class("renoise", "0", public)
end


--public functions

public.register_soundpack = function(ini_path)
	
	if not gksys.IsExist(ini_path) then
		return cerr("cannot register soundpack; path " .. tostring(ini_path) .. "was not valid")
	end
	
	if pack_registry[ini_path] then
		return cerr("cannot register soundpack; duplicate of existing entry")
	end
	
	local spack = {
		path = ini_path,
		name = rini("renoise", "name", ini_path),
		detail = rini("renoise", "detail", ini_path),
	}
	
	spack.load = gkrs("renoise", spack.path, "NO")
	
	if spack.name == "null" then
		return cerr("cannot register soundpack; path " .. tostring(ini_path) .. "was not valid")
	end
	
	cp("Registering new soundpack: " .. spack.name .. " at " .. ini_path)
	cp("	load state is " .. spack.load)
	
	table.insert(pack_registry, spack)
	pack_registry[ini_path] = true
	gkini.WriteString("renoise", ini_path, spack.load)
	gkini.WriteString("renoise", "sfx" .. tostring(#pack_registry), ini_path)
end



--self-forcing myself to develop front-end plugins in an interruptible non-blocking manner

public.get_soundpack_count = function()
	return #pack_registry
end

public.get_soundpack_item = function(id)
	return pack_registry[id]
end

public.edit_soundpack_item = function(id, entry, new_val)
	if not pack_registry[id] then
		return cerr("cannot edit soundpack entry; soundpack id invalid!")
	end
	
	local invalid_entries = {
		path = true,
		name = true,
	}
	
	if invalid_entries[entry] then
		return cerr("cannot edit soundpack entry; locked entry item")
	end
	
	pack_registry[id][entry] = new_val
end

public.save_config = function()
	--save config values
	for k, v in pairs(config) do
		gkini.WriteString("renoise", k, v)
	end
	
	--iterate over soundpacks, save load states
	local last_count = 0
	for k, v in ipairs(pack_registry) do
		gkini.WriteString("renoise", v.path, v.load)
		gkini.WriteString("renoise", "sfx" .. tostring(k), v.path)
		last_count = k
	end
	
	--remove entry after last valid
	gkini.WriteString("renoise", "sfx" .. tostring(last_count + 1), "")
end

public.set_config = function(cfg, val)
	--edits the config value. if no val is provided, returns current.
	if config[cfg] then
		if not val then
			return config[cfg]
		else
			config[cfg] = tostring(val)
		end
	end
end




--[[
	this runs at end-of-neoloader
	
	iterates over all soundpacks, building sfx change list, only applies latest entry afterwards
	
]]--

local apply_sfx = function(sound_table)
	--unloads the provided sound and applies the new effect
	local default = {
		--default entries for sound_table
		soundname = "null",
		filename = "null",
	}
	
	for k, v in pairs(sound_table) do
		default[k] = v
	end
	
	cp("	" .. default.soundname .. ": " .. default.filename)
	
	gksound.GKUnloadSound(default.soundname)
	
	local status 
	
	if default.filename == "NONE" then
		status = true
	else
		status = gksound.GKLoadSound(default)
	end
	
	cp("		status: " .. (status and "SUCCESS" or "FAILURE"))
	return status
end

local get_soundpack_tree_element = function(ini_path, sfx_title)
	--gets the table for a sound element in a soundpack
	local tree_object = {
		ini_path = ini_path,
		type = rini(sfx_title, "type", ini_path),
		soundname = rini(sfx_title, "soundname", ini_path),
		soundsearchpath = rini("renoise", "default_path", ini_path),
		soundfilename = rini(sfx_title, "filename", ini_path),
		samplerate = rini(sfx_title, "samplerate", ini_path),
		preload = rini(sfx_title, "preload", ini_path),
		stream = rini(sfx_title, "stream", ini_path),
		channels = rini(sfx_title, "channels", ini_path),
		biasdist = rini(sfx_title, "biasdist", ini_path),
		width = rini(sfx_title, "width", ini_path),
		loop = rini(sfx_title, "loop", ini_path),
	}
	
	local format = {
		ini_path = "string",
		type = "string",
		soundname = "string",
		soundsearchpath = "string",
		soundfilename = "string",
		samplerate = "number",
		preload = "boolean",
		stream = "boolean",
		channels = "number",
		biasdist = "number",
		width = "number",
		loop = "boolean",
	}
	
	--prune null elements
	for k, v in pairs(tree_object) do
		if v == "null" then
			tree_object[k] = nil
		elseif type(v) ~= format[k] then
			if format[k] == "string" then
				v = tostring(v)
			elseif format[k] == "number" then
				v = tonumber(v)
			elseif format[k] == "boolean" then
				v = (v == "true" and true or false)
			else
				v = nil
			end
			tree_object[k] = v
		end
	end
	
	if tree_object.soundfilename == "NONE" then
		tree_object.filename = "NONE"
	else
		tree_object.filename = tree_object.soundsearchpath .. tree_object.soundfilename
	end
	
	
	return tree_object
end

local get_soundpack_tree = function(ini_path)
	--gets the tables of a soundpack
	local tree = {
		--contains the media table related to each indexed header in the tree
	}
	
	local counter = 0
	while true do
		counter = counter + 1
		local sfx_title = rini("sfx", "sfx" .. tostring(counter), ini_path)
		
		if sfx_title == "null" then
			counter = counter - 1
			break
		end
		
		local sfx_object = get_soundpack_tree_element(ini_path, sfx_title)
		
		table.insert(tree, sfx_object)
	end
	
	return tree
end

local process_soundpack_list = function()
	--iterates over pack_registry (filled during lme loader), translates to flat table of sounds by name, then loads sounds. afx currently skipped.
	
	local sfx_list = {
		--list of media tables to be loaded
		--soundname = media,
	}
	
	local media_list = {
		--list of trees generated by parser
		--yes, I considered calling this table a 'forest'
		--currently gets discarded
	}
	
	
	cp("Processing all soundpacks...")
	
	for pack_id, pack_table in ipairs(pack_registry) do
		cp("	" .. pack_table.name)
		if pack_table.load == "YES" then
			local tree = get_soundpack_tree(pack_table.path)
			table.insert(media_list, tree)
			
			for tree_object, media_table in ipairs(tree) do
				if media_table.filename == "NONE" then
					sfx_list[media_table.soundname] = media_table
					cp("		" .. media_table.soundname .. " sent as mute signal")
				elseif gksys.IsExist(media_table.filename) then
					sfx_list[media_table.soundname] = media_table
					cp("		" .. media_table.soundname .. " sent to queue")
				else
					cp("		" .. media_table.soundname .. " skipped, file missing: " .. tostring(media_table.filename))
					cp("			media_table is " .. spickle(media_table))
				end
			end
		else
			cp("		skipped; not enabled")
		end
	end
	
	cp("Loading media...")
	
	--apply media
	for soundname, media in pairs(sfx_list) do
		apply_sfx(media)
	end
	
end

local load_config_list = function()
	--gets serial entries from config.ini
	local counter = 0
	while true do
		counter = counter + 1
		local next_entry = gkrs("renoise", "sfx" .. tostring(counter), "null")
		if next_entry == "null" then
			break
		end
		
		public.register_soundpack(next_entry)
	end
end

load_config_list()

RegisterEvent(process_soundpack_list, "PLUGINS_LOADED")





update_class()

lib.require({{name = "babel", version = "0"}}, babel_func)