--rngui, renoise lightweight gui
--manage renoise and available soundpacks

local cp = function(print_string)
	lib.log_error("[Renoise] " .. print_string, 1, "rngui", "0")
end

local cerr = function(err_string)
	lib.log_error("[Renoise] " .. err_string, 3, "rngui", "0")
	return false, err_string
end

local gkrs = gkini.ReadString

local renoise = lib.get_class("renoise", "0")
local helium = lib.get_class("helium", "0.4.2 -indev")



local public = {}



local bstr = function(id, val)
	return val
end

local update_class, babel, lang_key
local babel_func = function()
	babel = lib.get_class("babel", "0")
	lang_key = babel.register(lib.get_path("renoise", "0") .. "lang/", {'en', 'es', 'fr', 'pt'})
	
	bstr = function(id, val)
		return babel.fetch(lang_key, id, val)
	end
	
	public.add_translation = function(path, lang_code)
		babel.add_new_lang(ref_id, path, lang_code)
	end
	
	update_class()
end





update_class = function()
	local class = {
		CCD1 = true,
		open = nil,
		config = nil,
		description = bstr(7, "Renoise Lightweight Interface provides a minimalist front-end to managing soundpacks in the Renoise Core utility"),
		commands = {"/rngui"},
		manifest = {},
	}
	
	local post_update_options = {}
	
	class.smart_config = {
		title = "Renoise core module",
		cb = function(id, val)
			if id == "update_check" then
				Game.OpenWebBrowser("https://www.nexusmods.com/vendettaonline/mods/24")
			elseif id == "open_gui" then
				public.open()
			else
				--gkini.WriteString("renoise", id, tostring(val))
			end
		end,
		"open_gui",
		"spacer",
		"update_check",
		spacer = {
			type = "spacer",
		},
		open_gui = {
			type = "action",
			display = "Use the standard interface to manage renoise",
			align = "right",
			[1] = bstr(2, "Configure soundpacks") .. "...",
		},
		update_check = {
			type = "action",
			display = "",
			align = "right",
			[1] = bstr(3, "Check for updates") .. "...",
		},
	}
	
	--[[
		smart_config should be expanded with invdividual options to enable/disable unique soundpacks.
	]]--
	
	for k, v in pairs(post_update_options) do
		if type(k) == "number" then
			table.insert(class.smart_config, v)
		else
			class.smart_config[k] = v
		end
	end
	
	for k, v in pairs(class) do
		public[k] = v
	end
	
	lib.set_class("rngui", "0", public)
end







local create_interface = function(splist)
	--delayed until soundpack list is complete
	--splist: list of soundpack details
	
	local bghighlite_alternator = false
	
	local create_soundpack_line = function(spentry)
		--take a soundpack entry and create the toggle element
		local spe_toggle = iup.stationtoggle {
			title = "",
			value = spentry.load == "YES" and "ON" or "OFF",
			action = function(self)
				local new_value = self.value == "ON" and "YES" or "NO"
				renoise.edit_soundpack_item(spentry.position, "load", new_value)
			end,
		}
		
		local spe_name = iup.label {
			title = spentry.name,
		}
		
		local spe_path = iup.label {
			title = spentry.path,
			fgcolor = "128 128 128",
		}
		
		local spe_line = helium.frames.solidframe {
			bgcolor = bghighlite_alternator and "255 255 255" or "200 200 200",
			iup.hbox {
				spe_toggle,
				spe_name,
				iup.fill {},
				spe_path,
			},
		}
		
		return spe_line
	end
	
	local list_table = {}
	
	for k, v in ipairs(splist) do
		table.insert(list_table, create_soundpack_line(v))
	end
	
	local list_view = helium.vscroll(list_table)
	
	
	
	
	local close_action = function(iup_elem)
		renoise.save_config()
		HideDialog(iup.GetDialog(iup_elem))
	end
	
	local base_panel = iup.vbox {
		iup.hbox {
			iup.fill {
				size = "%2",
			},
			iup.vbox {
				alignment = "ACENTER",
				iup.multiline {
					size = "x%6",
					expand = "HORIZONTAL",
					value = bstr(4, "Toggle available soundpacks on and off here. Changes to audio will only occur after the game is fully restarted"),
					readonly = "YES",
				},
				iup.fill {
					size = "%2",
				},
				helium.frames.clearframe {
					size = "%75x" .. tostring(Font.Default * 6),
					list_view,
				},
				iup.fill {
					size = "%5",
				},
				iup.hbox {
					iup.stationbutton {
						title = bstr(5, "Save changes"),
						action = close_action,
					},
				},
			},
			iup.fill {
				size = "%2",
			},
		},
	}
	
	local diag = iup.dialog {
		bgcolor = "0 0 0 200 *",
		fullscreen = "YES",
		topmost = "YES",
		iup.vbox {
			iup.fill { },
			iup.hbox {
				iup.fill { },
				iup.stationdoubleframe {
					iup.hbox {
						iup.fill {
							size = "%5",
						},
						iup.vbox {
							iup.fill {
								size = "%2",
							},
							iup.label {
								title = "Renoise " .. bstr(6, "lightweight interface"),
							},
							helium.frames.clearframe {
								expand = "NO",
								base_panel,
							},
							iup.fill {
								size = "%2",
							},
						},
						iup.fill {
							size = "%5",
						},
					},
				},
				iup.fill { },
			},
			iup.fill { },
		},
	}
	
	helium.map_dialog(diag)
	
	ShowDialog(diag)
end






local interface_is_building_flag = false

public.open = function()
	if interface_is_building_flag then
		return
	end
	
	interface_is_building_flag = true
	
	local soundpack_list = {}
	local num_soundpacks = renoise.get_soundpack_count()
	
	local iter_timer = Timer()
	local get_entry_iterator
	get_entry_iterator = function()
		soundpack_list[num_soundpacks] = renoise.get_soundpack_item(num_soundpacks)
		soundpack_list[num_soundpacks].position = num_soundpacks
		num_soundpacks = num_soundpacks - 1
		if num_soundpacks > 0 then
			iter_timer:SetTimeout(1, get_entry_iterator)
		else
			create_interface(soundpack_list)
			interface_is_building_flag = false
		end
	end
	
	get_entry_iterator()
end



update_class()

lib.require({{name = "babel", version = "0"}}, babel_func)